import logging

from monopoly.config import PdfConfig, StatementConfig, TransactionConfig, settings
from monopoly.constants import (
    AccountType,
    BankNames,
    EncryptionIdentifier,
    MetadataIdentifier,
    StatementBalancePatterns,
    TransactionPatterns,
)

from ..base import BankBase

logger = logging.getLogger(__name__)


class Ocbc(BankBase):
    statement_config = StatementConfig(
        bank_name=BankNames.OCBC,
        account_type=AccountType.CREDIT,
        date_pattern=r"\d{2}\-\d{2}\-\d{4}",
        date_format=r"%d-%m-%Y",
        prev_balance_pattern=StatementBalancePatterns.OCBC,
    )

    transaction_config = TransactionConfig(
        pattern=TransactionPatterns.OCBC,
        date_format="%d/%m",
    )

    pdf_config = PdfConfig(
        passwords=settings.ocbc_pdf_passwords,
        page_range=(0, -2),
        page_bbox=(0, 0, 560, 750),
    )

    identifiers = [
        EncryptionIdentifier(
            pdf_version=1.4, algorithm=4, revision=4, length=128, permissions=-1036
        ),
        MetadataIdentifier(
            creator="pdfgen", producer="Streamline PDFGen for OCBC Group"
        ),
    ]



from datetime import datetime
from uuid import uuid4

from monopoly.config import StatementConfig


def generate_name(
    format_type: str, config: StatementConfig, statement_date: datetime
) -> str:
    bank_name = config.bank_name
    account_type = config.account_type
    year = statement_date.year
    month = statement_date.month

    filename = f"{bank_name}-{account_type}-{year}-{month:02d}"

    if format_type == "blob":
        return (
            f"bank_name={bank_name}/"
            f"account_type={account_type}/"
            f"year={year}/"
            f"month={month}/"
            f"{filename}-{uuid4().hex[0:8]}.csv"
        )

    if format_type == "file":
        return f"{filename}.csv"

    raise ValueError("Invalid format_type")



from monopoly.banks.base import BankBase
from monopoly.config import StatementConfig, TransactionConfig
from monopoly.constants import (
    AccountType,
    BankNames,
    MetadataIdentifier,
    SharedPatterns,
)


class ExampleBank(BankBase):
    """Dummy class to help with reading the example PDF statement"""

    statement_config = StatementConfig(
        bank_name=BankNames.EXAMPLE,
        account_type=AccountType.CREDIT,
        date_pattern=r"\d{2}\-\d{2}\-\d{4}",
        date_format=r"%d-%m-%Y",
        prev_balance_pattern=(
            r"(?P<description>LAST MONTH'S BALANCE?)\s+" + SharedPatterns.AMOUNT
        ),
    )

    transaction_config = TransactionConfig(
        pattern=(
            r"(?P<transaction_date>\d+/\d+)\s*"
            r"(?P<description>.*?)\s*" + SharedPatterns.AMOUNT
        ),
        date_format=r"%d/%m",
    )

    identifiers = [
        MetadataIdentifier(
            creator="Adobe Acrobat 23.3", producer="Adobe Acrobat Pro (64-bit)"
        )
    ]



import logging

from monopoly.config import PdfConfig, StatementConfig, TransactionConfig, settings
from monopoly.constants import (
    AccountType,
    BankNames,
    MetadataIdentifier,
    StatementBalancePatterns,
    TransactionPatterns,
)

from ..base import BankBase

logger = logging.getLogger(__name__)


class StandardChartered(BankBase):
    statement_config = StatementConfig(
        bank_name=BankNames.STANDARD_CHARTERED,
        account_type=AccountType.CREDIT,
        date_pattern=r"\d{2}\s\w+\s\d{4}",
        date_format=r"%d %B %Y",
        prev_balance_pattern=StatementBalancePatterns.STANDARD_CHARTERED,
    )

    transaction_config = TransactionConfig(
        pattern=TransactionPatterns.STANDARD_CHARTERED,
        date_format="%d %b",
    )

    pdf_config = PdfConfig(
        passwords=settings.standard_chartered_pdf_passwords,
        page_range=(0, -1),
        page_bbox=(0, 0, 580, 820),
    )

    identifiers = [
        MetadataIdentifier(
            title="eStatement",
            producer="iText",
        )
    ]



import logging
import re
from datetime import datetime
from functools import cached_property
from typing import Any, Optional

from pandas import DataFrame
from pydantic import field_validator, model_validator
from pydantic.dataclasses import dataclass
from pydantic_core import ArgsKwargs

from monopoly.config import StatementConfig, TransactionConfig
from monopoly.constants import StatementFields
from monopoly.pdf import PdfPage

logger = logging.getLogger(__name__)


# pylint: disable=bad-classmethod-argument
@dataclass
class Transaction:
    """
    Holds transaction data, validates the data, and
    performs various coercions like removing whitespace

    The class attributes are consistent with the enum
    values from StatementFields
    """

    transaction_date: str
    description: str
    amount: float
    suffix: Optional[str] = None

    @field_validator("description", mode="after")
    def remove_extra_whitespace(cls, value: str) -> str:
        return " ".join(value.split())

    @field_validator("amount", mode="before")
    def prepare_amount_for_float_coercion(cls, amount: str) -> str:
        """
        Replaces commas, whitespaces and parentheses in string representation of floats
        e.g.
            1,234.00 -> 1234.00
            (-10.00) -> -10.00
            (-1.56 ) -> -1.56
        """
        if isinstance(amount, str):
            return re.sub(r"[,)(\s]", "", amount)
        return amount

    @model_validator(mode="before")  # type: ignore
    def treat_parenthesis_enclosure_as_credit(self: ArgsKwargs | Any) -> "ArgsKwargs":
        """
        Treat amounts enclosed by parentheses (e.g. cashback) as a credit entry
        """
        if self.kwargs:
            amount: str = self.kwargs["amount"]
            if isinstance(amount, str):
                if amount.startswith("(") and amount.endswith(")"):
                    self.kwargs["suffix"] = "CR"
        return self

    @model_validator(mode="after")
    def convert_credit_amount_to_negative(self: "Transaction") -> "Transaction":
        """
        Converts transactions with a suffix of "CR" to negative
        """
        if self.suffix == "CR":
            self.amount = abs(self.amount)

        else:
            self.amount = -abs(self.amount)
        return self


@dataclass
class Statement:
    """
    A dataclass representation of a bank statement, containing
    the PDF pages (their raw text representation in a `list`),
    and specific config per bank.
    """

    pages: list[PdfPage]
    columns = [enum.value for enum in StatementFields]
    statement_config: StatementConfig
    transaction_config: TransactionConfig
    prev_month_balance = None

    @cached_property
    def transactions(self) -> list[Transaction]:
        transactions: list[Transaction] = []
        for page in self.pages:
            lines = self.process_lines(page)
            for i, line in enumerate(lines):
                transaction = self._process_line(line, lines, idx=i)
                if transaction:
                    transactions.append(transaction)
        transactions = self._add_prev_month_balance(transactions)
        return transactions

    def _process_line(
        self, line: str, lines: list[str], idx: int
    ) -> Transaction | None:
        if not self.prev_month_balance:
            if match := re.search(self.statement_config.prev_balance_pattern, line):
                self.prev_month_balance = match.groupdict()

        if match := re.search(self.transaction_config.pattern, line):
            transaction = Transaction(**match.groupdict())  # type: ignore

            # handle transactions that span multiple lines
            if self.transaction_config.multiline_transactions and idx < len(lines) - 1:
                next_line = lines[idx + 1]
                if not re.search(self.transaction_config.pattern, next_line):
                    transaction.description += " " + next_line
            return transaction
        return None

    def _add_prev_month_balance(
        self, transactions: list[Transaction]
    ) -> list[Transaction]:
        """
        Converts the previous month's statement balance into a transaction,
        if it exists in the statement.
        """
        if self.prev_month_balance:
            first_transaction_date = transactions[0].transaction_date
            prev_month_balance_transaction = Transaction(
                **self.prev_month_balance,  # type: ignore
                transaction_date=first_transaction_date,
            )
            transactions.insert(0, prev_month_balance_transaction)
        if not self.prev_month_balance:
            logger.debug("Unable to find previous month's balance")
        return transactions

    @staticmethod
    def process_lines(page: PdfPage) -> list:
        return [line.lstrip() for line in page.lines]

    @cached_property
    def raw_statement_date(self) -> str:
        config = self.statement_config
        logger.debug("Extracting statement date")
        first_page = self.pages[0]
        for line in first_page.lines:
            if match := re.findall(config.date_pattern, line):
                logger.debug("Statement date found")
                return match[0]
        raise ValueError("Statement date not found")

    @cached_property
    def statement_date(self):
        config = self.statement_config
        return datetime.strptime(self.raw_statement_date, config.date_format)

    @staticmethod
    def get_decimal_numbers(lines: list[str]) -> set[float]:
        """Returns all decimal numbers in a statement. This is used
        to perform a safety check, to make sure no transactions have been missed"""

        number_pattern = re.compile(r"[\d.,]+")
        decimal_pattern = re.compile(r"\d+\.\d+$")
        numbers = set()
        for line in lines:
            numbers.update(re.findall(number_pattern, line))
            numbers = {number.replace(",", "") for number in numbers}
            decimal_numbers = {
                float(number) for number in numbers if decimal_pattern.match(number)
            }
        return decimal_numbers

    @cached_property
    def df(self) -> DataFrame:
        return DataFrame(self.transactions, columns=self.columns)



from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Collection, Iterable, Optional

import click
from tqdm import tqdm

from monopoly.banks import auto_detect_bank


def process_statement(file: Path, output_directory: Optional[Path]):
    bank = auto_detect_bank(file)
    statement = bank.extract()
    transformed_df = bank.transform(statement)

    # saves processed statements to the same directory by default
    if not output_directory:
        output_directory = file.parent

    output_file = bank.load(transformed_df, statement, output_directory)
    return file.name, output_file.name


def run(input_files: Collection[Path], output_directory: Optional[Path] = None):
    with ProcessPoolExecutor() as executor:
        results = list(
            tqdm(
                executor.map(
                    process_statement,
                    input_files,
                    [output_directory] * len(input_files),
                ),
                total=len(input_files),
                desc="Processing statements",
                ncols=80,
                bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt}",
            )
        )

    click.echo(click.style(f"{len(input_files)} statement(s) processed", bold=True))
    for raw_statement, processed_statement in sorted(results):
        click.echo(f"{raw_statement} -> {processed_statement}")


def get_statement_paths(files: Iterable[Path]) -> set[Path]:
    matched_files = set()
    for path in files:
        if path.is_file() and str(path).endswith(".pdf"):
            matched_files.add(path)

        if path.is_dir():
            matched_files |= get_statement_paths(path.iterdir())

    return matched_files


@click.command()
@click.argument(
    "files",
    nargs=-1,
    type=click.Path(exists=True, allow_dash=True, resolve_path=True, path_type=Path),
)
@click.option(
    "-o",
    "--output",
    type=click.Path(exists=True, allow_dash=True, resolve_path=True, path_type=Path),
    help="Specify output folder",
)
@click.pass_context
def monopoly(ctx: click.Context, files: list[Path], output: Path):
    """
    Monopoly helps convert your bank statements from PDF to CSV.

    A file or directory can be passed in via the FILES argument
    """
    if files:
        matched_files = get_statement_paths(files)

        if matched_files:
            run(matched_files, output)
            ctx.exit(0)

        else:
            click.echo(
                click.style("Could not find .pdf files", fg="yellow", bold=True),
                err=True,
            )
            ctx.exit(1)
    else:
        show_welcome_message()


def show_welcome_message():
    art = r"""
     __  __                               _
    |  \/  | ___  _ __   ___  _ __   ___ | |_   _
    | |\/| |/ _ \| '_ \ / _ \| '_ \ / _ \| | | | |
    | |  | | (_) | | | | (_) | |_) | (_) | | |_| |
    |_|  |_|\___/|_| |_|\___/| .__/ \___/|_|\__, |
                             |_|            |___/
    """

    click.echo(art)
    message = "Monopoly helps convert your bank statements from PDF to CSV"
    click.echo(message)
    commands = [
        (
            "monopoly .",
            "process all statements nested in current dir",
        ),
        (
            "monopoly path/to/file.pdf",
            "process a specific statement in a directory",
        ),
        (
            "monopoly . --output <dir>",
            "saves all results to a specific directory",
        ),
        (
            "monopoly --help",
            "show more options and other usage information",
        ),
    ]
    margin = max(len(cmd) for cmd, _ in commands)
    for cmd, desc in commands:
        styled_cmd = click.style(
            text=f"{cmd}{' ' * (margin - len(cmd))}",
            fg="white",
            bg="bright_black",
            bold=True,
        )
        click.echo(message=f"    {styled_cmd} {desc}")
    click.echo()



import logging

from monopoly.config import PdfConfig, StatementConfig, TransactionConfig, settings
from monopoly.constants import (
    AccountType,
    BankNames,
    MetadataIdentifier,
    StatementBalancePatterns,
    TransactionPatterns,
)

from ..base import BankBase

logger = logging.getLogger(__name__)


class Citibank(BankBase):
    statement_config = StatementConfig(
        bank_name=BankNames.CITIBANK,
        account_type=AccountType.CREDIT,
        date_pattern=r"Statement\sDate\s+(.*)",
        date_format=r"%B %d, %Y",
        prev_balance_pattern=StatementBalancePatterns.CITIBANK,
    )

    transaction_config = TransactionConfig(
        pattern=TransactionPatterns.CITIBANK,
        date_format="%d %b",
    )

    pdf_config = PdfConfig(
        passwords=settings.citibank_pdf_passwords,
        page_bbox=(20, 0, 595, 840),
        page_range=(0, -3),
    )

    identifiers = [
        MetadataIdentifier(
            creator="Ricoh Americas Corporation, AFP2PDF",
            producer="Ricoh Americas Corporation, AFP2PDF",
        )
    ]



import logging

from monopoly.config import PdfConfig, StatementConfig, TransactionConfig
from monopoly.constants import (
    AccountType,
    BankNames,
    EncryptionIdentifier,
    MetadataIdentifier,
    StatementBalancePatterns,
    TransactionPatterns,
)

from ..base import BankBase

logger = logging.getLogger(__name__)


class Dbs(BankBase):
    statement_config = StatementConfig(
        bank_name=BankNames.DBS,
        account_type=AccountType.CREDIT,
        date_pattern=r"(\d{2}\s[A-Za-z]{3}\s\d{4})",
        date_format=r"%d %b %Y",
        prev_balance_pattern=StatementBalancePatterns.DBS,
    )

    transaction_config = TransactionConfig(
        pattern=TransactionPatterns.DBS,
        date_format="%d %b",
    )

    pdf_config = PdfConfig(page_range=(0, -1))

    identifiers = [
        EncryptionIdentifier(
            pdf_version=1.4, algorithm=2, revision=3, length=128, permissions=-1852
        ),
        MetadataIdentifier(creator="Quadient CXM AG"),
    ]
